import logging
from ibmsecurity.utilities import tools
import ibmsecurity.isam.web.reverse_proxy.junctions_server as junctions_server
import json
from ibmsecurity.isam.web.reverse_proxy.junctions_config import server_fields
from ibmsecurity.utilities.tools import jsonSortedListEncoder

try:
    basestring
except NameError:
    basestring = (str, bytes)

# does not work anymore in Python 3.11+
logger = logging.getLogger(__name__)


# URI for this module
uri = "/wga/reverseproxy"
requires_modules = ["wga"]
requires_version = None



def get_all(isamAppliance, reverseproxy_id, detailed=None, check_mode=False, force=False, warnings=[]):
    """
    Retrieving a list of standard and virtual junctions

    :param isamAppliance:
    :param reverseproxy_id:
    :param detailed: Set to True if you want a detailed junction list (new in v10.0.4)
    :param check_mode:
    :param force:
    :return:
    """
    if detailed and tools.version_compare(isamAppliance.facts["version"], "10.0.4") >= 0:
        try:
            returnValue = isamAppliance.invoke_get("Retrieving a list of standard and virtual junctions",
                                        "{0}/{1}/junctions?detailed=true".format(uri, reverseproxy_id),
                                        requires_modules=requires_modules,
                                        requires_version=requires_version)
        except:
            warnings.append("Detailed retrieval of junctions failed unexpectedly.  Falling back to simple version.")
            returnValue = isamAppliance.invoke_get("Retrieving a list of standard and virtual junctions (fallback)",
                                    "{0}/{1}/junctions".format(uri, reverseproxy_id),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version,
                                    warnings=warnings
            )
            returnValue['warnings'] = warnings
        return returnValue
    else:
      #ignore detailed for older versions
      return isamAppliance.invoke_get("Retrieving a list of standard and virtual junctions",
                                    "{0}/{1}/junctions".format(uri, reverseproxy_id),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, reverseproxy_id, junctionname, check_mode=False, force=False, warnings=[]):
    """
    Retrieving the parameters for a single standard or virtual junction

    :param isamAppliance:
    :param reverseproxy_id:
    :param junctionname:
    :param check_mode:
    :param force:
    :param warnings
    :return:
    """
    logger = isamAppliance.logger
    ret_obj = isamAppliance.invoke_get("Retrieving the parameters for a single standard or virtual junction",
                                       "{0}/{1}/junctions?junctions_id={2}".format(uri, reverseproxy_id,
                                                                                   junctionname),
                                       requires_modules=requires_modules,
                                       requires_version=requires_version,
                                       warnings=warnings)
    # servers are provided as a single string, here we parse it out into a list + dict
    servers = []
    if tools.version_compare(isamAppliance.facts["version"], "9.0.1.0") > 0:
        srv_separator = '#'
    else:
        srv_separator = '&'
    srvs = ret_obj['data']['servers'].split(srv_separator)
    logger.debug("Servers in raw string: {0}".format(ret_obj['data']['servers']))
    logger.debug("Number of servers in junction: {0}".format(len(srvs)))
    for srv in srvs:
        logger.debug("Parsing Server: {0}".format(srv))
        server = {}
        for s in srv.split(';'):
            if s != '':
                kv = s.split('!')
                server[kv[0]] = kv[1]
        servers.append(server)

    ret_obj['data']['servers'] = servers
    return ret_obj


def _check(isamAppliance, reverseproxy_id, junctionname, currentJunctions=None):
    """ CurrentJunctions is the output of get_all.
        This avoids constantly having to call the get_all function.
    """
    if currentJunctions is None:
        ret_obj = get_all(isamAppliance, reverseproxy_id)
    else:
        ret_obj = currentJunctions

    for jct in ret_obj['data']:
        if jct['id'] == junctionname:
            return True

    return False


def add(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port, junction_type="tcp",
        virtual_hostname=None, server_dn=None, query_contents=None, stateful_junction=None, case_sensitive_url=None,
        windows_style_url=None, https_port=None, http_port=None, proxy_hostname=None, proxy_port=None,
        sms_environment=None, vhost_label=None, junction_hard_limit=None, junction_soft_limit=None,
        basic_auth_mode=None, tfim_sso=None, remote_http_header=None, preserve_cookie=None, cookie_include_path=None,
        transparent_path_junction=None, mutual_auth=None, insert_session_cookies=None, request_encoding=None,
        enable_basic_auth=None, key_label=None, gso_resource_group=None, junction_cookie_javascript_block=None,
        client_ip_http=None, version_two_cookies=None, ltpa_keyfile=None, authz_rules=None, fsso_config_file=None,
        username=None, password=None, server_uuid=None, local_ip=None, ltpa_keyfile_password=None,
        delegation_support=None, scripting_support=None, insert_ltpa_cookies=None, check_mode=False, force=False,
        http2_junction=None, http2_proxy=None, sni_name=None, description=None,
        priority=None, server_cn=None, silent=None,
        case_insensitive_url=None,
        servers=None,
        warnings=[]):
    """
    Creating a standard or virtual junction

    :param isamAppliance:
    :param reverseproxy_id:
    :param junction_point:
    :param server_hostname:
    :param server_port:
    :param junction_type:
    :param virtual_hostname:
    :param server_dn:
    :param query_contents:
    :param stateful_junction:
    :param case_sensitive_url:
    :param case_insensitive_url: #v10.0.6
    :param windows_style_url:
    :param https_port:
    :param http_port:
    :param proxy_hostname:
    :param proxy_port:
    :param sms_environment:
    :param vhost_label:
    :param junction_hard_limit:
    :param junction_soft_limit:
    :param basic_auth_mode:
    :param tfim_sso:
    :param remote_http_header:
    :param preserve_cookie:
    :param cookie_include_path:
    :param transparent_path_junction:
    :param mutual_auth:
    :param insert_session_cookies:
    :param request_encoding:
    :param enable_basic_auth:
    :param key_label:
    :param gso_resource_group:
    :param junction_cookie_javascript_block:
    :param client_ip_http:
    :param version_two_cookies:
    :param ltpa_keyfile:
    :param authz_rules:
    :param fsso_config_file:
    :param username:
    :param password:
    :param server_uuid:
    :param local_ip:
    :param ltpa_keyfile_password:
    :param delegation_support:
    :param scripting_support:
    :param insert_ltpa_cookies:
    :param check_mode:
    :param force:
    :param http2_junction:
    :param http2_proxy:
    :param sni_name:
    :param description:
    :param priority:
    :param server_cn:
    :param silent:
    :return:
    """
    # See if it's a virtual or standard junction
    isVirtualJunction = True
    logger = isamAppliance.logger
    if junction_point[:1] == '/':
        isVirtualJunction = False
    if force is True or _check(isamAppliance, reverseproxy_id, junction_point) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            # Create a simple json with just the main junction attributes
            jct_json = {
                "junction_point": junction_point,
                "junction_type": junction_type.lower(),
                "server_hostname": server_hostname,
                "server_port": server_port,
                "force": force
            }
            # Add attributes that have been supplied... otherwise skip them.
            if junction_hard_limit is not None:
                jct_json["junction_hard_limit"] = junction_hard_limit
            if junction_soft_limit is not None:
                jct_json["junction_soft_limit"] = junction_soft_limit
            if basic_auth_mode is not None:
                jct_json["basic_auth_mode"] = basic_auth_mode
            if tfim_sso is not None:
                jct_json["tfim_sso"] = tfim_sso
            if remote_http_header is not None:
                jct_json["remote_http_header"] = remote_http_header
            if stateful_junction is not None:
                jct_json["stateful_junction"] = stateful_junction
            if not isVirtualJunction and preserve_cookie:
                jct_json["preserve_cookie"] = preserve_cookie
            if not isVirtualJunction and cookie_include_path:
                jct_json["cookie_include_path"] = cookie_include_path
            if not isVirtualJunction and transparent_path_junction:
                jct_json["transparent_path_junction"] = transparent_path_junction
            if mutual_auth is not None:
                jct_json["mutual_auth"] = mutual_auth
            if insert_ltpa_cookies is not None:
                jct_json["insert_ltpa_cookies"] = insert_ltpa_cookies
            if insert_session_cookies is not None:
                jct_json["insert_session_cookies"] = insert_session_cookies
            if request_encoding is not None:
                jct_json["request_encoding"] = request_encoding
            if enable_basic_auth is not None:
                jct_json["enable_basic_auth"] = enable_basic_auth
            if key_label is not None:
                jct_json["key_label"] = key_label
            if gso_resource_group is not None:
                jct_json["gso_resource_group"] = gso_resource_group
            if junction_cookie_javascript_block is not None and junction_cookie_javascript_block != '':
                jct_json["junction_cookie_javascript_block"] = junction_cookie_javascript_block
            if client_ip_http is not None:
                jct_json["client_ip_http"] = client_ip_http
            if version_two_cookies is not None:
                jct_json["version_two_cookies"] = version_two_cookies
            if ltpa_keyfile is not None:
                jct_json["ltpa_keyfile"] = ltpa_keyfile
            if authz_rules is not None:
                jct_json["authz_rules"] = authz_rules
            if fsso_config_file is not None:
                jct_json["fsso_config_file"] = fsso_config_file
            if username is not None:
                jct_json["username"] = username
            if password is not None:
                jct_json["password"] = password
            if server_uuid is not None:
                jct_json["server_uuid"] = server_uuid
            if virtual_hostname is not None:
                jct_json["virtual_hostname"] = virtual_hostname
            if server_dn is not None:
                jct_json["server_dn"] = server_dn
            if local_ip is not None:
                jct_json["local_ip"] = local_ip
            if query_contents is not None:
                jct_json["query_contents"] = query_contents
            if tools.version_compare(isamAppliance.facts["version"], "10.0.6.0") >= 0:
                # If no case_insensitive_url is passed, we take the old one and invert it.
                # Who thinks it's a good idea to make changes in an API like this ?
                if case_insensitive_url is not None:
                    jct_json["case_insensitive_url"] = case_insensitive_url
                elif case_sensitive_url is not None:
                    if case_sensitive_url.lower() == 'no':
                        jct_json["case_insensitive_url"] = 'yes'
                    else:
                        jct_json["case_insensitive_url"] = 'no' # default
                else:
                    jct_json["case_insensitive_url"] = 'no'
            elif case_sensitive_url is not None:
                jct_json['case_sensitive_url'] = case_sensitive_url
            if windows_style_url is not None:
                jct_json["windows_style_url"] = windows_style_url
            if ltpa_keyfile_password is not None:
                jct_json["ltpa_keyfile_password"] = ltpa_keyfile_password
            if https_port is not None:
                jct_json["https_port"] = https_port
            if http_port is not None:
                jct_json["http_port"] = http_port
            if proxy_hostname is not None:
                jct_json["proxy_hostname"] = proxy_hostname
            if proxy_port is not None:
                jct_json["proxy_port"] = proxy_port
            if isVirtualJunction and sms_environment is not None:
                jct_json["sms_environment"] = sms_environment
            if isVirtualJunction and vhost_label is not None:
                jct_json["vhost_label"] = vhost_label
            if delegation_support is not None:
                jct_json["delegation_support"] = delegation_support
            if scripting_support is not None:
                jct_json["scripting_support"] = scripting_support
            if http2_junction is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, http2_junction: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring http2_junction for this call.".format(
                            isamAppliance.facts["version"], http2_junction))
                else:
                    jct_json["http2_junction"] = http2_junction
            if http2_proxy is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, http2_proxy: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring http2_proxy for this call.".format(
                            isamAppliance.facts["version"], http2_proxy))
                else:
                    jct_json['http2_proxy'] = http2_proxy
            if sni_name is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, sni_name: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring sni_name for this call.".format(
                            isamAppliance.facts["version"], sni_name))
                else:
                    jct_json['sni_name'] = sni_name
            if description is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.7.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, description: {1} is not supported. Needs 9.0.7.0 or higher. Ignoring description for this call.".format(
                            isamAppliance.facts["version"], description))
                else:
                    jct_json['description'] = description
            if priority is not None:
                if tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, priority: {1} is not supported. Needs 10.0.2.0 or higher. Ignoring priority for this call.".format(
                            isamAppliance.facts["version"], priority))
                else:
                    jct_json['priority'] = priority
            else:
                if tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") >= 0:
                    warnings.append(
                        "Appliance at version: {0}, priority is required".format(
                            isamAppliance.facts["version"]))
                    jct_json['priority'] = "9"
                else:
                    priority = None
            if server_cn is not None:
                if tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, server_cn: {1} is not supported. Needs 10.0.2.0 or higher. Ignoring server_cn for this call.".format(
                            isamAppliance.facts["version"], server_cn))
                else:
                    jct_json['server_cn'] = server_cn
            if isVirtualJunction and silent:
                jct_json['silent'] = silent
            return isamAppliance.invoke_post(
                "Creating a standard or virtual junction",
                "{0}/{1}/junctions".format(uri, reverseproxy_id), jct_json,
                requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, reverseproxy_id, junctionname, check_mode=False, force=False):
    """
    Deleting a standard or virtual junction

    :param isamAppliance:
    :param reverseproxy_id:
    :param junctionname:
    :param check_mode:
    :param force:
    :return:
    """
    if force is True or _check(isamAppliance, reverseproxy_id, junctionname) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a standard or virtual junction",
                "{0}/{1}/junctions?junctions_id={2}".format(uri, reverseproxy_id, junctionname),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def set(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port, junction_type="tcp",
        virtual_hostname=None, server_dn=None, query_contents=None, stateful_junction=None, case_sensitive_url=None,
        windows_style_url=None, https_port=None, http_port=None, proxy_hostname=None, proxy_port=None,
        sms_environment=None, vhost_label=None, junction_hard_limit=None, junction_soft_limit=None,
        basic_auth_mode=None, tfim_sso=None, remote_http_header=None, preserve_cookie=None, cookie_include_path=None,
        transparent_path_junction=None, mutual_auth=None, insert_session_cookies=None, request_encoding=None,
        enable_basic_auth=None, key_label=None, gso_resource_group=None, junction_cookie_javascript_block=None,
        client_ip_http=None, version_two_cookies=None, ltpa_keyfile=None, authz_rules=None, fsso_config_file=None,
        username=None, password=None, server_uuid=None, local_ip=None, ltpa_keyfile_password=None,
        delegation_support=None, scripting_support=None, insert_ltpa_cookies=None, check_mode=False, force=False,
        http2_junction=None, http2_proxy=None, sni_name=None, description=None,
        priority=None, server_cn=None, silent=None, case_insensitive_url=None, servers=None, warnings=[]):
    """
    Setting a standard or virtual junction - compares with existing junction and replaces if changes are detected
    """
    add_required = False
    logger = isamAppliance.logger
    # See if it's a virtual or standard junction
    isVirtualJunction = True
    if junction_point[:1] == '/':
        isVirtualJunction = False
        logger.debug("Junction: {0} is a standard junction".format(junction_point))
    if force is False:
        # Check if record exists
        logger.debug("Check if the junction exists.")
        if _check(isamAppliance, reverseproxy_id, junction_point):
            logger.debug("Junction exists. Compare junction details.")
            ret_obj = get(isamAppliance, reverseproxy_id=reverseproxy_id, junctionname=junction_point, warnings=warnings)
            exist_jct = ret_obj['data']
            jct_json = {
                'junction_point': junction_point,
                'junction_type': junction_type.lower(),
                'server_hostname': server_hostname,
                'server_port': server_port
            }
            logger.debug("See if the backend junction server matches any on the junction. Look for just one match.")
            srvs = ret_obj['data']['servers']
            srvs_len = len(srvs)

            if not junction_server_exists(isamAppliance, srvs=srvs,
                                          server_hostname=server_hostname,
                                          server_port=server_port,
                                          case_sensitive_url=case_sensitive_url,
                                          isVirtualJunction=isVirtualJunction,
                                          http_port= http_port,
                                          local_ip=local_ip,
                                          query_contents=query_contents,
                                          server_dn=server_dn,
                                          server_uuid=server_uuid,
                                          virtual_hostname=virtual_hostname,
                                          windows_style_url=windows_style_url,
                                          priority=priority,
                                          server_cn=server_cn,
                                          case_insensitive_url=case_insensitive_url):
                add_required = True
            elif not add_required:
                if tools.version_compare(isamAppliance.facts["version"], "10.0.6.0") >= 0:
                    # If no case_insensitive_url is passed, we take the old one and invert it.
                    # Who thinks it's a good idea to make changes in an API like this ?
                    if case_insensitive_url is not None:
                        jct_json["case_insensitive_url"] = case_insensitive_url
                    elif case_sensitive_url is not None:
                        if case_sensitive_url.lower() == 'no':
                            jct_json["case_insensitive_url"] = 'yes'
                        else:
                            jct_json["case_insensitive_url"] = 'no'  # default
                elif case_sensitive_url is not None:
                    jct_json['case_sensitive_url'] = case_sensitive_url

                if authz_rules is None:
                    jct_json['authz_rules'] = 'no'
                else:
                    jct_json['authz_rules'] = authz_rules
                if basic_auth_mode is None:
                    jct_json['basic_auth_mode'] = 'filter'
                else:
                    jct_json['basic_auth_mode'] = basic_auth_mode
                if client_ip_http is None or client_ip_http.lower() == 'no':
                    jct_json['client_ip_http'] = 'do not insert'
                elif client_ip_http.lower() == 'yes':
                    jct_json['client_ip_http'] = 'insert'
                else:
                    jct_json['client_ip_http'] = client_ip_http
                if not isVirtualJunction:
                  logger.debug("Only for standard junctions - {0}.".format(virtual_hostname))
                  if cookie_include_path is None:
                      jct_json['cookie_include_path'] = 'no'
                  else:
                      jct_json['cookie_include_path'] = cookie_include_path
                if isVirtualJunction and sms_environment:
                    jct_json['sms_environment'] = sms_environment
                if delegation_support is None:
                    jct_json['delegation_support'] = 'no'
                else:
                    jct_json['delegation_support'] = delegation_support
                if fsso_config_file is None or fsso_config_file == '':
                    jct_json['fsso_config_file'] = 'disabled'
                else:
                    jct_json['fsso_config_file'] = fsso_config_file
                if insert_session_cookies is None:
                    jct_json['insert_session_cookies'] = 'no'
                else:
                    jct_json['insert_session_cookies'] = insert_session_cookies
                if junction_hard_limit is None:
                    jct_json['junction_hard_limit'] = '0 - using global value'
                else:
                    jct_json['junction_hard_limit'] = str(junction_hard_limit)
                if junction_soft_limit is None:
                    jct_json['junction_soft_limit'] = '0 - using global value'
                else:
                    jct_json['junction_soft_limit'] = str(junction_soft_limit)

                if mutual_auth is None:
                    jct_json['mutual_auth'] = 'no'
                else:
                    jct_json['mutual_auth'] = mutual_auth
                if not isVirtualJunction:
                    if preserve_cookie is None:
                        jct_json['preserve_cookie'] = 'no'
                    else:
                        jct_json['preserve_cookie'] = preserve_cookie

                if remote_http_header is None or remote_http_header == []:
                    logger.debug("\nSetting remote_http_header to do not insert")
                    if not isVirtualJunction:
                        jct_json['remote_http_header'] = 'do not insert'
                else:
                   jct_json['remote_http_header'] = [_word.replace('_', '-') for _word in
                                                     list(remote_http_header)]
                # To allow for multiple header values to be sorted during compare convert retrieved data into array
                if request_encoding is None:
                    jct_json['request_encoding'] = 'utf8_uri' # utf8_bin, utf8_uri, lcp_bin, and lcp_uri
                else:
                    jct_json['request_encoding'] = request_encoding
                if scripting_support is None:
                    jct_json['scripting_support'] = 'no'
                else:
                    jct_json['scripting_support'] = scripting_support
                if stateful_junction is None:
                    jct_json['stateful_junction'] = 'no'
                else:
                    jct_json['stateful_junction'] = stateful_junction
                if tfim_sso is None:
                    jct_json['tfim_sso'] = 'no'
                else:
                    jct_json['tfim_sso'] = tfim_sso
                if transparent_path_junction is None:
                    jct_json['transparent_path_junction'] = 'no'
                else:
                    jct_json['transparent_path_junction'] = transparent_path_junction
                if isVirtualJunction:
                   logger.debug("Only for virtual junctions - virtual hostname {0}.".format(virtual_hostname))

                   if virtual_hostname:
                       jct_json['virtual_junction_hostname'] = virtual_hostname
                   else:
                       if jct_json['server_port'] in ['80', '443', 80, 443]:
                           jct_json['virtual_junction_hostname'] = jct_json['server_hostname']
                       else:
                           jct_json['virtual_junction_hostname'] = jct_json['server_hostname'] + ":" +  jct_json['server_port']
                if http2_junction is not None and http2_junction != "no":
                    if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                        warnings.append(
                            "Appliance at version: {0}, http2_junction: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring http2_junction for this call.".format(
                                isamAppliance.facts["version"], http2_junction))
                        http2_junction = None
                    else:
                        jct_json['http2_junction'] = http2_junction
                        #if 'http2_junction' not in exist_jct:
                        #    exist_jct['http2_junction'] = jct_json['http2_junction']
                if http2_proxy is not None and http2_proxy != "no":
                    if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                        warnings.append(
                            "Appliance at version: {0}, http2_proxy: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring http2_proxy for this call.".format(
                                isamAppliance.facts["version"], http2_proxy))
                        http2_proxy = None
                    else:
                        jct_json['http2_proxy'] = http2_proxy
                        #if 'http2_proxy' not in exist_jct:
                        #    exist_jct['http2_proxy'] = jct_json['http2_proxy']
                if sni_name is not None:
                    if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                        warnings.append(
                            "Appliance at version: {0}, sni_name: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring sni_name for this call.".format(
                                isamAppliance.facts["version"], sni_name))
                        sni_name = None
                    else:
                        jct_json['sni_name'] = sni_name
                if description is not None:
                    if tools.version_compare(isamAppliance.facts["version"], "9.0.7.0") < 0:
                        warnings.append(
                            "Appliance at version: {0}, description: {1} is not supported. Needs 9.0.7.0 or higher. Ignoring description for this call.".format(
                                isamAppliance.facts["version"], description))
                        description = None
                    else:
                        jct_json['description'] = description
                if isVirtualJunction and silent:
                    jct_json['silent'] = silent

            if junction_exists(isamAppliance, exist_jct, jct_json):
                add_required = False
            else:
                add_required = True
            if add_required and srvs_len > 1:
                warnings.append(
                    "Junction will be replaced. Existing multiple servers #{0} will be overwritten. Please re-add as needed.".format(
                        srvs_len))
        else:
            add_required = True

    if force or add_required:
        # Junction force add will replace the junction, no need for delete (force set to True as a result)
        return add(isamAppliance=isamAppliance, reverseproxy_id=reverseproxy_id, junction_point=junction_point,
                   server_hostname=server_hostname, server_port=server_port, junction_type=junction_type,
                   virtual_hostname=virtual_hostname, server_dn=server_dn, query_contents=query_contents,
                   stateful_junction=stateful_junction, case_sensitive_url=case_sensitive_url,
                   windows_style_url=windows_style_url, https_port=https_port, http_port=http_port,
                   proxy_hostname=proxy_hostname, proxy_port=proxy_port, sms_environment=sms_environment,
                   vhost_label=vhost_label, junction_hard_limit=junction_hard_limit,
                   junction_soft_limit=junction_soft_limit, basic_auth_mode=basic_auth_mode, tfim_sso=tfim_sso,
                   remote_http_header=remote_http_header, preserve_cookie=preserve_cookie,
                   cookie_include_path=cookie_include_path, transparent_path_junction=transparent_path_junction,
                   mutual_auth=mutual_auth, insert_session_cookies=insert_session_cookies,
                   request_encoding=request_encoding, enable_basic_auth=enable_basic_auth, key_label=key_label,
                   gso_resource_group=gso_resource_group,
                   junction_cookie_javascript_block=junction_cookie_javascript_block, client_ip_http=client_ip_http,
                   version_two_cookies=version_two_cookies, ltpa_keyfile=ltpa_keyfile, authz_rules=authz_rules,
                   fsso_config_file=fsso_config_file, username=username, password=password, server_uuid=server_uuid,
                   local_ip=local_ip, ltpa_keyfile_password=ltpa_keyfile_password,
                   delegation_support=delegation_support, scripting_support=scripting_support,
                   insert_ltpa_cookies=insert_ltpa_cookies, check_mode=check_mode, force=True,
                   http2_junction=http2_junction, http2_proxy=http2_proxy, sni_name=sni_name, description=description,
                   priority=priority, server_cn=server_cn, silent=silent,
                   case_insensitive_url=case_insensitive_url,
                   warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def compare(isamAppliance1, isamAppliance2, reverseproxy_id, reverseproxy_id2=None):
    """
    Compare list of junctions in a given reverse proxy between 2 appliances
    """
    if reverseproxy_id2 is None or reverseproxy_id2 == '':
        reverseproxy_id2 = reverseproxy_id
    ret_obj1 = get_all(isamAppliance1, reverseproxy_id)
    ret_obj2 = get_all(isamAppliance2, reverseproxy_id2)

    for jct in ret_obj1['data']:
        ret_obj = get(isamAppliance1, reverseproxy_id, jct['id'])
        del ret_obj['data']['active_worker_threads']
        for srv in ret_obj['data']['servers']:
            del srv['current_requests']
            del srv['operation_state']
            del srv['server_state']
            if ret_obj['data']['stateful_junction'] == 'no':
                del srv['server_uuid']
            del srv['total_requests']
        jct['details'] = ret_obj['data']

    for jct in ret_obj2['data']:
        ret_obj = get(isamAppliance2, reverseproxy_id2, jct['id'])
        del ret_obj['data']['active_worker_threads']
        for srv in ret_obj['data']['servers']:
            del srv['current_requests']
            del srv['operation_state']
            del srv['server_state']
            if ret_obj['data']['stateful_junction'] == 'no':
                del srv['server_uuid']
            del srv['total_requests']
        jct['details'] = ret_obj['data']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['active_worker_threads', 'servers/current_requests',
                                                                'servers/operation_state', 'servers/server_state',
                                                                'servers/server_uuid', 'servers/total_requests'])

def set_all(isamAppliance, reverseproxy_id: str, junctions: list=[], check_mode=False, force=False, warnings=[]):
    """
    Set junctions with all the servers
    The input is a list of junction objects, that can be passed to the `set` function
    The list of junctions is first compared to the output of `get_all`, so we only need to update junctions that are changed.

    :param isamAppliance:
    :param reverseproxy_id:
    :param junctions: List of junctions to set.  This is a list of dicts, with each dict representing a junction
    :param check_mode:
    :param force:
    :return:
    """
    logger = isamAppliance.logger

    currentJunctions = get_all(isamAppliance, reverseproxy_id=reverseproxy_id, detailed=True)

    __markChanged = False # use this bool to indicate if there's been a change or not.

    if currentJunctions['rc'] == 0:
        logger.debug(f"\nCurrent junctions:\n{currentJunctions}")
    else:
        # no junctions exists
        logger.debug("No junctions exist yet.  Create them all.")
        # Force create - There are no junctions yet
        # use expansion
        __markChanged = True
        for j in junctions:
            j.pop('isVirtualJunction', None)
            __firstserver = j.get('servers', [''])[0]
            for _field in list(server_fields.keys()):
                if __firstserver.get(_field, None) is not None and j.get(_field, None) is None:
                    # only use server_fields if they are not defined on junction level
                        logger.debug(f"{_field} from {__firstserver.get(_field)} copied to junction level")
                        j[_field] = __firstserver.get(_field, None)
            if len(j.get('servers', [''])) > 1:
                __servers = j.get('servers', [''])[1:]
            else:
                __servers = None
            j.pop('servers', None)
            j['force'] = True  # Force create
            set(isamAppliance, reverseproxy_id, **j)
            # Also add servers (if servers[] has more than 1 item)
            if __servers is not None:
                for s in __servers:
                    for _field, kval in server_fields.items():
                        if s.get(_field, None) is not None:
                            j[_field] = s.get(_field, None)
                    junctions_server.set(isamAppliance, reverseproxy_id, **j)
    # Compare the junctions and the currentJunctions.
    for j in junctions:
        logger.debug(f"Processing junction: {j['junction_point']}")

        _checkUpdate = False
        j['isVirtualJunction'] = True
        if j['junction_point'][:1] == '/':
            j['isVirtualJunction'] = False
        if j.get('junction_soft_limit', None) is None:
            j['junction_soft_limit'] = '0 - using global value'
        else:
           j['junction_soft_limit'] = str( j.get('junction_soft_limit', None))
        if j.get('junction_hard_limit', None) is None:
            j['junction_hard_limit'] = '0 - using global value'
        else:
           j['junction_hard_limit'] = str( j.get('junction_hard_limit', None))
        if j.get('client_ip_http', None) is None or j.get('client_ip_http', '').lower() == 'no':
            j['client_ip_http'] = 'do not insert'
        elif j.get('client_ip_http', '').lower() == 'yes':
            j['client_ip_http'] = 'insert'
        if j.get('junction_type', None) is not None:
            j['junction_type'] = j.get('junction_type', '').lower() # if junction_type is empty, rest api will fail anyway
        # update remote http header logic here
        if j.get('remote_http_header', None) is None:
            logger.debug("No remote http header")
        elif isinstance(j.get('remote_http_header', None), list):
            j['remote_http_header'] = [_word.replace('_', '-') for _word in
                                              j.get('remote_http_header', None)]
        else:
            j['remote_http_header'] = [j.get('remote_http_header', '')]

        # check that we have the required fields, if not, get them from the first server (if that exists)
        __firstserver = j.get('servers', [''])[0]
        for _field in list(server_fields.keys()):
            if __firstserver.get(_field, None) is not None and j.get(_field, None) is None:
                # only use server_fields if they are not defined on junction level
                logger.debug(f"{_field} from {__firstserver.get('server_hostname')} copied to junction level")
                j[_field] = __firstserver.get(_field, None)
        if len(j.get('servers', [''])) > 1:
            __servers = j.get('servers', [''])[1:]

        for c in currentJunctions['data']:
            servers = []
            if tools.version_compare(isamAppliance.facts["version"], "9.0.1.0") > 0:
                srv_separator = '#'
            else:
                srv_separator = '&'

            if c.get('junction_point', None) == j['junction_point']:
                if c.get('servers', None) is not None:
                    if type(c.get('servers', [])) is str:
                        srvs = c['servers'].split(srv_separator)
                        logger.debug("Servers in raw string: {0}".format(c['servers']))
                        logger.debug("Number of servers in junction: {0}".format(len(srvs)))
                        for srv in srvs:
                            logger.debug("Parsing Server: {0}".format(srv))
                            server = {}
                            for s in srv.split(';'):
                                if s != '':
                                    kv = s.split('!')
                                    server[kv[0]] = kv[1]
                            servers.append(server)
                        c['servers'] = servers
                else:
                    c['servers'] = []

                logger.debug(f"The junction at {j['junction_point']} already exists.")
                _checkUpdate = dict(c)
                break
            elif c.get('id', None) == j['junction_point']:
                logger.debug(f"The junction at {j['junction_point']} already exists (simple syntax)")
                warnings.append(f"Had to use simple get syntax unexpectedly for {j['junction_point']}")
                _checkUpdate = get(isamAppliance, reverseproxy_id, j['junction_point'], check_mode=False, force=False, warnings=warnings)
                _checkUpdate = _checkUpdate.get('data', _checkUpdate)
                if _checkUpdate.get('servers', None) is not None:
                    if type(_checkUpdate.get('servers', [])) is str:
                        srvs = _checkUpdate['servers'].split(srv_separator)
                        logger.debug("Servers in raw string: {0}".format(_checkUpdate['servers']))
                        logger.debug("Number of servers in junction: {0}".format(len(srvs)))
                        for srv in srvs:
                            logger.debug("Parsing Server: {0}".format(srv))
                            server = {}
                            for s in srv.split(';'):
                                if s != '':
                                    kv = s.split('!')
                                    server[kv[0]] = kv[1]
                            servers.append(server)
                        _checkUpdate['servers'] = servers
                else:
                    _checkUpdate['servers'] = []
                break

        if _checkUpdate:
            __firstserver = j.get('servers', [''])[0]
            for _field in list(server_fields.keys()):
                if __firstserver.get(_field, None) is not None and j.get(_field, None) is None:
                    # only use server_fields if they are not defined on junction level
                    logger.debug(f"{_field} from {__firstserver.get('server_hostname')} copied to junction level")
                    j[_field] = __firstserver.get(_field, None)
            if not junction_exists(isamAppliance, _checkUpdate, j):
                __markChanged = True
                # Run set()
                logger.debug("\n\nUpdate junction\n\n")
                warnings.append(f"Instance {reverseproxy_id}: Updating junction {j['junction_point']}")
                j['force'] = True  # force create
                j['warnings'] = warnings
                j.pop('isVirtualJunction', None)
                __result = set(isamAppliance, reverseproxy_id, **j)
                logger.debug(f"Adding servers")
                # Also add servers (if servers[] has more than 1 item)
                if len(j.get('servers', [''])) > 1:
                    j.pop('servers', None)
                    for s in __servers:
                        for _field, kval in server_fields.items():
                            if s.get(_field, None) is not None:
                                j[_field] = s.get(_field, None)
                        junctions_server.set(isamAppliance, reverseproxy_id, **j)
            else:
                logger.debug(f"\n\n{reverseproxy_id}: Junction {j.get('junction_point','')} does not need updating\n\n")
                warnings.append(f"Instance {reverseproxy_id}: Junction {j.get('junction_point','')} does not need updating")
        else:
            # Force create - this junction does not exist yet
            j.pop('isVirtualJunction', None)
            j['force'] = True # force create
            __markChanged = True
            j['warnings'] = warnings
            __firstserver = j.get('servers', [''])[0]
            for _field in list(server_fields.keys()):
                if __firstserver.get(_field, None) is not None and j.get(_field, None) is None:
                    # only use server_fields if they are not defined on junction level
                        logger.debug(f"{_field} from {__firstserver.get('server_hostname')} copied to junction level")
                        j[_field] = __firstserver.get(_field, None)
            logger.debug(f"Creating new junction with {j}")
            set(isamAppliance, reverseproxy_id, **j)
            logger.debug(f"Adding servers")
            # Also add servers (if servers[] has more than 1 item)
            if len(j.get('servers', [''])) > 1:
                __servers = j.get('servers', [''])[1:]
                j.pop('servers', None)
                for s in __servers:
                    for _field, kval in server_fields.items():
                        if s.get(_field, None) is not None:
                            j[_field] = s.get(_field, None)
                    junctions_server.set(isamAppliance, reverseproxy_id, **j)
    if __markChanged:
        return isamAppliance.create_return_object(changed=True, warnings=warnings)
    else:
        return isamAppliance.create_return_object(warnings=warnings)


def junction_server_exists(isamAppliance, srvs, server_hostname: str, server_port, case_sensitive_url: str='yes',
                         isVirtualJunction=False, http_port=None, local_ip=None,
                         query_contents=None, server_dn=None,
                         server_uuid=None, virtual_hostname=None, windows_style_url=None, priority=None, server_cn=None, case_insensitive_url: str='no', **kwargs) -> bool:
  server_found = False
  for srv in srvs:
    if srv['server_hostname'] == server_hostname and str(srv['server_port']) == str(server_port):
        logger.debug("Matched a server - {0}.".format(srv))
        server_found = True
        server_json = {
            'server_hostname': server_hostname,
            'server_port': str(server_port)
        }

        if tools.version_compare(isamAppliance.facts["version"], "10.0.6.0") >= 0:
            # If no case_insensitive_url is passed, we take the old one and invert it.
            # Who thinks it's a good idea to make changes in an API like this ?
            if case_insensitive_url is not None:
                server_json["case_insensitive_url"] = case_insensitive_url
            elif case_sensitive_url is not None:
                if case_sensitive_url.lower() == 'no':
                    server_json["case_insensitive_url"] = 'yes'
                else:
                    server_json["case_insensitive_url"] = 'no'  # default
        elif case_sensitive_url is not None:
            server_json['case_sensitive_url'] = case_sensitive_url
        else:
            server_json['case_sensitive_url'] = 'yes'

        if http_port is None:
            server_json['http_port'] = str(server_port)
        else:
            server_json['http_port'] = str(http_port)
        if local_ip is None:
            server_json['local_ip'] = ''
        else:
            server_json['local_ip'] = local_ip
        if query_contents is None or query_contents == '':
            server_json['query_content_url'] = '/cgi-bin/query_contents'
        else:
            server_json['query_content_url'] = query_contents
        if server_dn is None:
            server_json['server_dn'] = ''
        else:
            server_json['server_dn'] = server_dn
        if server_uuid is not None:
            server_json['server_uuid'] = server_uuid
        else:
            # Server UUID gets generated if not specified
            srv.pop('server_uuid', None)
        if not isVirtualJunction:
            if virtual_hostname is not None:
                logger.debug("Only for standard junctions - {0}.".format(virtual_hostname))
                server_json['virtual_junction_hostname'] = virtual_hostname
            else:
                if server_json['server_port'] in ['80', 80]:
                    server_json['virtual_junction_hostname'] = server_json['server_hostname']
                else:
                    server_json['virtual_junction_hostname'] = server_json['server_hostname'] + ":" + server_json[
                        'server_port']
            if 'virtual_junction_hostname' not in srv:
                # this is not in the returned servers object for virtual host junctions, it's in the junction's object
                if virtual_hostname is not None:
                    srv['virtual_junction_hostname'] = virtual_hostname
                else:
                    if srv['server_port'] in ['80', '443', 80, 443]:
                        srv['virtual_junction_hostname'] = srv['server_hostname']
                    else:
                        srv['virtual_junction_hostname'] = 'fuckoff' # srv['server_hostname'] + ":" + srv['server_port']
        if windows_style_url is None:
            server_json['windows_style_url'] = 'no'
        else:
            server_json['windows_style_url'] = windows_style_url
        # v10.0.2
        if tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") >= 0:
            if priority is None:
                server_json['priority'] = '9'
            else:
                server_json['priority'] = str(priority)
            if server_cn is None:
                server_json['server_cn'] = ''
            else:
                server_json['server_cn'] = server_cn

        srv = {k: v for k, v in srv.items() if k in server_json.keys()}

        current_servers = json.dumps(srv, skipkeys=True, sort_keys=True)
        new_servers = json.dumps(server_json, skipkeys=True, sort_keys=True)
        if current_servers != new_servers:
            logger.debug("\n\nServers are found to be different. See following JSON for difference.\n\n")
        else:
            logger.debug("\n\nServers are the same.  See comparison below.\n\n")
        logger.debug(f"\nNew Server JSON: {new_servers}")
        logger.debug(f"\nOld Server JSON: {current_servers}")
        break
  return server_found

def junction_exists(isamAppliance, exist_jct, new_j):
    # perform a comparison.  we receive the actual current junction as input here
    __srvs = exist_jct.get('servers', None)
    new_j.pop('isVirtualJunction', None)
    logger.debug(f"\n\nServers in junction {new_j['junction_point']}:\n{__srvs}")
    __result = True
    if __srvs is not None and not junction_server_exists(isamAppliance, __srvs, **new_j):
        __result = False
    if __result:
        # ok we still need to compare
        # exist_jct = dict(_checkUpdate)
        new_jct = dict(new_j)
        # remove the server fields - this has already been compared
        for _field, kval in server_fields.items():
            new_jct.pop(_field, None)
        # junction_type
        if exist_jct.get('junction_type', None) is not None:
            exist_jct['junction_type'] = exist_jct.get('junction_type', '').lower()
        # insert_ltpa_cookies
        if exist_jct.get('insert_ltpa_cookies', None) is None:
            exist_jct['insert_ltpa_cookies'] = 'no'
        # sms_environment
        if exist_jct.get('sms_environment', None) is None:
            exist_jct['sms_environment'] = ""
        # vhost_label
        if exist_jct.get('vhost_label', None) is None:
            exist_jct['vhost_label'] = ""
        # request_encoding utf8_bin, utf8_uri, lcp_bin, and lcp_uri.
        __re = exist_jct.get('request_encoding', None)
        if __re is not None:
            if __re == 'UTF-8, URI Encoded':
                exist_jct['request_encoding'] = 'utf8_uri'
            elif __re == 'UTF-8, Binary':
                exist_jct['request_encoding'] = 'utf8_bin'
            elif __re == 'Local Code Page, Binary':
                exist_jct['request_encoding'] = 'lcp_bin'
            elif __re == 'Local Code Page, URI Encoded':
                exist_jct['request_encoding'] = 'lcp_uri'

        # To allow for multiple header values to be sorted during compare convert retrieved data into array
        __rehh = exist_jct.get('remote_http_header', None)
        if __rehh is not None:
            if __rehh.startswith('insert - '):
                exist_jct['remote_http_header'] = [_word.replace('_', '-') for _word in (__rehh[9:]).split(' ')]
            # now see if it's actually 'all' - to do that, compare with string ["iv-creds", "iv-groups", "iv-user"]
            __nrehh = json.dumps(exist_jct.get('remote_http_header', None), skipkeys=True, sort_keys=True, cls=jsonSortedListEncoder)
            logger.debug(f"Sorted string content of remote_http_header {__nrehh}")
            if __nrehh == '["iv-creds", "iv-groups", "iv-user"]':
                exist_jct['remote_http_header'] = ['all']
            if __nrehh.replace('"', '') == "do not insert":
                exist_jct['remote_http_header'] = []

        # scripting support default
        __scripting_support = exist_jct.get('scripting_support', None)
        if __scripting_support is None:
            exist_jct['scripting_support'] = 'no'

        # scripting support default
        __transparent_path_junction = exist_jct.get('transparent_path_junction', None)
        if __transparent_path_junction is None:
            exist_jct['transparent_path_junction'] = 'no'

        # basic_auth_mode - filter (default), ignore, supply, gso.
        __bamode = exist_jct.get('basic_auth_mode', None)
        if __bamode is not None:
            # GSO
            if __bamode == "use GSO":
                exist_jct['basic_auth_mode'] = 'gso'

        # Remove servers field for comparing
        new_jct.pop("servers", None)
        # only compare values that are in the new request
        # This does not (always) compare values correctly where you just remove the key.  In that case, you'd have to change a different key as well (eg. description)
        exist_jct = {k: v for k, v in exist_jct.items() if k in new_jct.keys()}

        newJSON = json.dumps(new_jct, skipkeys=True, sort_keys=True, cls=jsonSortedListEncoder)
        logger.debug(f"\nSorted Desired  Junction {new_j['junction_point']}:\n\n {newJSON}\n")

        oldJSON = json.dumps(exist_jct, skipkeys=True, sort_keys=True, cls=jsonSortedListEncoder)
        logger.debug(f"\nSorted Current  Junction {exist_jct.get('junction_point', '')}:\n\n {oldJSON}\n")

        if newJSON != oldJSON:
            logger.debug("Junctions are found to be different. See JSON for difference.")
            __result = False
    return __result




import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/kerberos/config/realms/"
requires_modules = ['wga']
requires_version = None


def get_all(isamAppliance, recursive='yes', includeValuesInLine='yes', addParent='yes', check_mode=False, force=False):
    """
    Retrieve Kerberos Configuration: Realms
    """
    return isamAppliance.invoke_get("Retrieve Kerberos Configuration: Realms",
                                    "{0}{1}".format(uri, tools.create_query_string(recursive=recursive,
                                                                                   includeValuesInLine=includeValuesInLine,
                                                                                   addParent=addParent)),
                                    requires_modules=requires_modules, requires_version=requires_version)


def add(isamAppliance, realm, check_mode=False, force=False):
    """
    Creating a Kerberos realm (Subsection)
    """
    ret_obj = search(isamAppliance=isamAppliance, realm=realm)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Creating a Kerberos realm",
                "{0}".format(uri),
                {
                    "subsection": realm
                })

    return isamAppliance.create_return_object()


def delete(isamAppliance, realm, check_mode=False, force=False):
    """
    Deleting a Kerberos realm
    """
    ret_obj = search(isamAppliance=isamAppliance, realm=realm)

    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a Kerberos realm",
                "{0}{1}".format(uri, realm))

    return isamAppliance.create_return_object()


def get(isamAppliance, realm, check_mode=False, force=False):
    """
    Retrieve a specific Kerberos Configuration realm
    """
    return isamAppliance.invoke_get("Retrieve a specific Kerberos realm details",
                                    "{0}{1}".format(uri, realm),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def search(isamAppliance, realm, check_mode=False, force=False):
    """
    Search kerberos realm by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()
    return_obj["warnings"] = ret_obj["warnings"]

    for obj in ret_obj['data']:
        if obj['name'] == realm:
            logger.info("Found Kerberos realm {0} id: {1}".format(realm, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def compare(isamAppliance1, isamAppliance2):
    """
    Compare kerberos configuration realms between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/override-configs"
requires_modules = ["mga", "federation"]
requires_version = None
mga_uri = "/mga/advanced_config/files"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of configuration properties
    """
    return isamAppliance.invoke_get("Retrieve a list of configuration properties",
                                    "/iam/access/v8/override-configs", requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, key, check_mode=False, force=False):
    """
    Retrieve a specific configuration property
    """
    ret_obj = search(isamAppliance, key=key)
    id = ret_obj['data']

    if id == {}:
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieve a specific configuration property",
                                        "/iam/access/v8/override-configs/{0}".format(id),
                                        requires_modules=requires_modules, requires_version=requires_version)


def update(isamAppliance, key, value, sensitive, check_mode=False, force=False):
    """
    Update a configuration property
    """
    id, matches = None, None
    if force is False:
        id, matches = _check(isamAppliance, key, value, sensitive)

    if force is True or (matches is False and id is not None):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a configuration property",
                "/iam/access/v8/override-configs/{0}".format(id),
                {
                    'value': value,
                    'sensitive': sensitive
                }, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def update_all(isamAppliance, values, check_mode=False, force=False):
    """
    Update a list of configuration properties

    Unlike the REST API use key values instead of id
    """
    matches, vals = None, values
    if force is False:
        matches, vals = _check_values(isamAppliance, values)

    if force is True or matches is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a list of configuration properties",
                "/iam/access/v8/override-configs",
                vals, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def get_files(isamAppliance, path, check_mode=False, force=False):
    """
    Retrieving a list of configuration files
    :param isamAppliance:
    :param check_mode:
    :param force:
    :return:
    """
    return isamAppliance.invoke_get("Retrieving a list of configuration files",
                                    "{0}/{1}".format(mga_uri, path),
                                    requires_modules=requires_modules, requires_version="10.0.0")


def get_paths(isamAppliance, check_mode=False, force=False):
    """
    Retrieving a list of configuration file component paths
    :param isamAppliance:
    :param check_mode:
    :param force:
    :return:
    """
    return isamAppliance.invoke_get("Retrieving a list of configuration file component paths",
                                    "{0}".format(mga_uri),
                                    requires_modules=requires_modules, requires_version="10.0.0")


def get_contents(isamAppliance, path, filename, check_mode=False, force=False):
    """
    Retrieving a list of configuration file component paths
    :param isamAppliance:
    :param check_mode:
    :param force:
    :return:
    """
    return isamAppliance.invoke_get("Retrieving a list of configuration file component paths",
                                    "{0}/{1}/{2}".format(mga_uri, path, filename),
                                    requires_modules=requires_modules, requires_version="10.0.0")


def export_file(isamAppliance, path, filename, file_path, check_mode=False, force=False):
    """
    Exporting a configuration file
    :param isamAppliance:
    :param check_mode:
    :param force:
    :return:
    """
    import os.path

    if force is True or (os.path.exists(file_path) is False):
        return isamAppliance.invoke_get_file("Exporting a configuration file",
                                             "{0}/{1}/{2}?export=true".format(mga_uri, path, filename),
                                             file_path,
                                             requires_modules=requires_modules, requires_version="10.0.0")

    return isamAppliance.create_return_object(warnings=["File {0} already exists".format(file_path)])


def update_file_contents(isamAppliance, path, filename, contents, check_mode=False, force=False):
    """
    Update the contents of a configuration file
    :param isamAppliance:
    :param check_mode:
    :param force:
    :return:
    """
    update, warnings = _check_contents(isamAppliance, path, filename, contents)

    if force is True or update is True:
        return isamAppliance.invoke_put("Update the contents of a configuration file",
                                        "{0}/{1}/{2}".format(mga_uri, path, filename),
                                        {
                                            'contents': contents,
                                            'type': 'file'
                                        },
                                        requires_modules=requires_modules, requires_version="10.0.0")
    return isamAppliance.create_return_object(warnings=warnings)


def import_file(isamAppliance, path, filename, file_path, check_mode=False, force=False):
    """
    Upload a new configuration file
    :param isamAppliance:
    :param check_mode:
    :param force:
    :return:
    """
    update, warnings = _check_import(isamAppliance, path, filename, file_path)

    if force is True or update is True:
        return isamAppliance.invoke_put_files("Upload a new configuration file",
                                              "{0}/{1}/{2}?uiCalled=true".format(mga_uri, path, filename),
                                              [
                                                  {
                                                      'file_formfield': 'file',
                                                      'filename': file_path,
                                                      'mimetype': 'application/octet-stream'
                                                  }
                                              ],
                                              {
                                                  "force": "yes",
                                                  "type": "file"
                                              },
                                              requires_modules=requires_modules, requires_version="10.0.0")
    return isamAppliance.create_return_object(warnings=warnings)


def _check_contents(isamAppliance, path, filename, contents):
    ret_obj = get_contents(isamAppliance, path, filename)
    if 'contents' in ret_obj['data']:
        if (ret_obj['data']['contents']) == contents:
            return False, ret_obj['warnings']
        else:
            return True, ret_obj['warnings']
    else:
        return False, ret_obj['warnings']


def _check_import(isamAppliance, path, filename, file_path):
    ret_obj = get_contents(isamAppliance, path, filename)

    import os
    if os.path.exists(file_path) is False:
        warnings = file_path + " does not exist."
        return False, [warnings]

    with open(file_path, 'r') as infile:
        input_contents = infile.read()

    if 'contents' in ret_obj['data']:
        if (ret_obj['data']['contents']) == input_contents:
            return False, ret_obj['warnings']
        else:
            return True, ret_obj['warnings']
    else:
        return False, ret_obj['warnings']


def _check(isamAppliance, key, value, sensitive):
    """
    Check and see if given key value matches existing values to be updated and also
    extract the id to use instead of the key
    """
    id, matches = None, None
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['key'] == key:
            id = obj['id']
            if obj['value'] == value and obj['sensitive'] == sensitive:
                matches = True
            else:
                matches = False
            break

    return id, matches


def _check_values(isamAppliance, values):
    """
    Filter out values to be updated, and evaluate if an update needs to happen

    Note: Key values not found will be skipped!
    """
    matches, vals = True, []
    ret_obj = get_all(isamAppliance)

    for new_obj in values:
        for obj in ret_obj['data']:
            if obj['key'] == new_obj['key']:
                if obj['value'] != new_obj['value'] or obj['sensitive'] != new_obj['sensitive']:
                    matches = False
                    vals.append({'id': obj['id'], 'value': new_obj['value'], 'sensitive': new_obj['sensitive']})
                break

    logger.debug(vals)

    return matches, vals


def search(isamAppliance, key):
    """
    Get the ID for a given key value
    """
    ret_obj = get_all(isamAppliance)

    ret_obj_new = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['key'] == key:
            ret_obj_new['data'] = obj['id']
            break

    return ret_obj_new


def compare(isamAppliance1, isamAppliance2):
    """
    Compare AAC Advanced configuration between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

requires_modules = ["mga", "federation"]
requires_version = None


def get_all(isamAppliance, check_mode=False, force=False, ignore_error=False):
    """
    Retrieving the current runtime template files directory contents
    """
    return isamAppliance.invoke_get("Retrieving the current runtime template files directory contents",
                                    "/mga/template_files?recursive=yes", ignore_error=ignore_error,
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, path, recursive=None, check_mode=False, force=False):
    """
    Retrieving the current runtime template files directory contents
    """
    return isamAppliance.invoke_get("Retrieving the current runtime template files directory contents",
                                    "/mga/template_files/{}/{}".format(path,
                                                                       ibmsecurity.utilities.tools.create_query_string(
                                                                           recursive=recursive)),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def _check(isamAppliance, id):
    ret_obj = get_all(isamAppliance, ignore_error=True)

    if (ret_obj['rc'] != 404 or ret_obj['rc'] != 400):
        return _parse_id(ret_obj['data'], id)
    else:
        return None


def _parse_id(contents, dir_name):
    """
    Recursively parse and find the id for a given directory

    :param contents:
    :param dir_name:
    :return id:
    """
    try:
        split_dir = dir_name.split('/', 1)
        cur_dir = split_dir[0]
        rest_dir = split_dir[1]
    except:
        rest_dir = ''
    for dir in contents:
        if dir['name'] == cur_dir and dir['type'] == 'Directory':
            if rest_dir == '':
                return dir['id']
            else:
                if len(dir['children']) == 0:
                    return None
                else:
                    return _parse_id(dir['children'], rest_dir)

    return None


def create(isamAppliance, path, name, check_mode=False, force=False):
    """
    Creating a directory in the runtime template files directory

    :param isamAppliance:
    :param id:
    :param name:
    :param check_mode:
    :param force:
    :return:
    """
    warnings = []
    id = path + "/" + name
    check_dir = _check(isamAppliance, id)
    if check_dir != None:
        warnings.append("Directory {0} exists in path {1}. Ignoring create.".format(name, path))

    if force is True or check_dir == None:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                "Creating a directory in the runtime template files directory",
                "/mga/template_files/{0}".format(path),
                {
                    'dir_name': name,
                    'type': 'dir'
                }, requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Deleting a file or directory in the runtime template files directory

    :param isamAppliance:
    :param id:
    :param check_mode:
    :param force:
    :return:
    """
    warnings = []
    check_dir = _check(isamAppliance, id)
    if check_dir == None:
        warnings.append("Directory {0} does not exist. Ignoring delete.".format(id))

    if force is True or check_dir != None:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a directory in the runtime template files directory",
                "/mga/template_files/{0}".format(id), requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def rename(isamAppliance, id, new_name, check_mode=False, force=False):
    """
    Deleting a file or directory in the runtime template files directory

    :param isamAppliance:
    :param id:
    :param new_name:
    :param check_mode:
    :param force:
    :return:
    """
    warnings = []

    try:
        split_dir = id.split('/')
        path = split_dir[:-1]
        new_path = '/'.join([str(x) for x in path]) + "/" + new_name
    except:
        logger.info("New path can't be build from id: {0} and new_name: {1}.".format(id, new_name))

    if force is False:
        new_dir_id = _check(isamAppliance, new_path)
        dir_id = _check(isamAppliance, id)

    if new_dir_id != None:
        warnings.append("Directory {0} does already exist. Ignoring renameing.".format(new_path))

    if force is True or (dir_id != None and new_dir_id == None):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Renaming a directory in the runtime template files directory",
                "/mga/template_files/{0}".format(id),
                {
                    'new_name': new_name,
                    'type': 'directory'
                }, requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def _remove_version(contents):
    """
    Recursively parse and remove version

    :param contents:
    :return:
    """
    for obj in contents:
        try:
            del obj['version']
            if len(obj['children']) != 0:
                _remove_version(obj['children'])
        except:
            pass
    return


def compare(isamAppliance1, isamAppliance2, path):
    ret_obj1 = get(isamAppliance1, path)
    ret_obj2 = get(isamAppliance2, path)

    _remove_version(ret_obj1['data'])
    _remove_version(ret_obj2['data'])

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['version'])



import logging

logger = logging.getLogger(__name__)
uri = "/isam/pdadmin"
version = "v1"

def get(isamAppliance, admin_id, admin_pwd, admin_domain='Default', **kwargs):
    """
    Retrieve a list of POPs that match a specific pop_attribute_name or pop_attribute_value
    (not useful at all)
    :param str admin_id: policy administrator, typically sec_master
    :param str admin_pwd:
    :param str admin_domain: the webseal domain, defaults to 'Default' if not supplied
    :param str pop_attribute_name: required if pop_attribute_value is not defined
    :param str pop_attribute_value: required if pop_attribute_name is not defined
    :param bool check_mode: not used
    :param bool force: not used
    :return: the returned data
    :rtype: object
    """

    input_args = {
        "admin_id": admin_id,
        "admin_pwd": admin_pwd
    }
    if admin_domain:
        input_args["admin_domain"] = admin_domain

    for k, v in kwargs.items():
        if k == 'check_mode':
            continue
        if k == 'force':
            continue
        input_args[k] = v

    ret_obj = isamAppliance.invoke_post("Retrieve a list of POPs",
                                        f"{uri}/poplistext/{version}",
                                        ignore_error=True,
                                        data=input_args)
    ret_obj['changed'] = False
    if ret_obj['rc'] == 404:
        logger.info(f"No POPs found matching your arguments {input_args}")
        return isamAppliance.create_return_object()
    return ret_obj

def retrieve(isamAppliance, admin_id, admin_pwd, pop_name, admin_domain='Default', **kwargs):
    """
    Retrieve a specific POP (show)
    """
    ret_obj = isamAppliance.invoke_post("Retrieve a specific POP",
                                        f"{uri}/popshowext/{version}",
                                        ignore_error=True,
                                        data={
                                            "admin_id": admin_id,
                                            "admin_pwd": admin_pwd,
                                            "pop_name": pop_name,
                                            "admin_domain": admin_domain
                                        })
    ret_obj['changed'] = False
    if ret_obj['rc'] == 404:
        logger.info(f"No POP {pop_name} found in {admin_domain}")
        return isamAppliance.create_return_object()
    return ret_obj

def get_objects(isamAppliance, admin_id, admin_pwd, **kwargs):
    """
    Retrieve a list of protected objects
    :param str admin_id: policy administrator, typically sec_master
    :param str admin_pwd:
    :param str admin_domain: the webseal domain, defaults to 'Default' if not supplied
    :param str object: optional
    :param str pop_name: the name of the pop, optional
    :param str pop_attribute_name: optional
    :param str pop_attribute_value: optional
    :param bool check_mode: not used
    :param bool force: not used
    :return: the returned data
    :rtype: object
    """
    input_args = {
        "admin_id": admin_id,
        "admin_pwd": admin_pwd
    }
    for k, v in kwargs.items():
        if k == 'check_mode':
            continue
        if k == 'force':
            continue
        input_args[k] = v

    ret_obj = isamAppliance.invoke_post("Retrieve a list of protected objects",
                                        f"{uri}/popfindext/{version}",
                                        ignore_error=True,
                                        data=input_args)
    if ret_obj['rc'] == 404:
        logger.info(f"No pops found for your arguments {input_args}")
        return isamAppliance.create_return_object()
    ret_obj['changed'] = False
    return ret_obj

#alias for function get_objects
get_all = get_objects

def compare(isamAppliance1, isamAppliance2, isamUser, admin_domain='Default', **kwargs):
    """
    Compare URL Mapping between two appliances
    TODO: THIS IS NOT READY
    """
    ret_obj1 = get_all(isamAppliance1, isamUser=isamUser, admin_domain=admin_domain)
    ret_obj2 = get_all(isamAppliance2, isamUser=isamUser, admin_domain=admin_domain)

    for obj in ret_obj1['data']:
        del obj['version']
        ret_obj = get(isamAppliance1, pop_name=obj['id'], isamUser=isamUser, admin_domain=admin_domain)
        obj['script'] = ret_obj['data']['contents']
    for obj in ret_obj2['data']:
        del obj['version']
        ret_obj = get(isamAppliance2, pop_name=obj['id'], isamUser=isamUser, admin_domain=admin_domain)
        obj['script'] = ret_obj['data']['contents']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['version'])
